
LogisticReg <- function(exp, obs, exp_lats, exp_lons, obs_lats, obs_lons, target_grid, int_method = NULL, 
                  points = NULL, method_point_interp = NULL, lat_dim = "lat", lon_dim = "lon",
                  sdate_dim = "sdate", member_dim = "member", source_file = NULL, region = NULL, 
                  loocv = FALSE, ncores = 1) {

  #-----------------------------------
  # Checkings
  #-----------------------------------
  if (!inherits(target_grid, 'character')) {
    stop("Parameter 'target_grid' must be of the class 'character'")
  }

  if (!is.null(int_method) & !inherits(int_method, 'character')) {
    stop("Parameter 'int_method' must be of the class 'character'")
  }

  if (!is.null(method_point_interp) & !inherits(method_point_interp, 'character')) {
    stop("Parameter 'method_point_interp' must be of the class 'character'")
  }

  if (!inherits(lat_dim, 'character')) {
    stop("Parameter 'lat_dim' must be of the class 'character'")
  }

  if (!inherits(lon_dim, 'character')) {
    stop("Parameter 'lon_dim' must be of the class 'character'")
  }

  if (!inherits(sdate_dim, 'character')) {
    stop("Parameter 'sdate_dim' must be of the class 'character'")
  }

  if (!inherits(member_dim, 'character')) {
    stop("Parameter 'member_dim' must be of the class 'character'")
  }

  if (!is.null(source_file) & !inherits(source_file, 'character')) {
    stop("Parameter 'source_file' must be of the class 'character'")
  }

  if (!inherits(loocv, 'logical')) {
    stop("Parameter 'loocv' must be set to TRUE or FALSE")
  }

  if (is.na(match(lon_dim, names(dim(exp)))) | is.na(match(lon_dim, names(dim(obs))))) {
    stop("Missing longitude dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'lon_dim'")
  }

  if (is.na(match(lat_dim, names(dim(exp)))) | is.na(match(lat_dim, names(dim(obs))))) {
    stop("Missing latitude dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'lat_dim'")
  }

  if (is.na(match(sdate_dim, names(dim(exp)))) | is.na(match(sdate_dim, names(dim(obs))))) {
    stop("Missing start date dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'sdate_dim'")
  }

  if (is.na(match(member_dim, names(dim(exp))))) {
    stop("Missing member dimension in 'exp' and/or 'obs', or does not match the parameter ",
         "'member_dim'")
  }

  if (!is.null(points) & (is.null(source_file))) {
    stop("No source files found. One source file for exp must be provided in the parameter 'source_file'.")
  }

  if (!is.null(points) & is.null(method_point_interp)) {
    stop("Please provide the interpolation method to interpolate gridded data to point locations ",
         "through the parameter 'method_point_interp'.")
  }

  if (is.null(region)) {
    warning("The borders of the downscaling region have not been provided. Assuming the four borders ",
            "of the downscaling region are defined by the first and last elements of the parametres ",
            "'obs_lats' and 'obs_lons'.")
    region <- c(obs_lons[1], obs_lons[length(obs_lons)], obs_lats[1], obs_lats[length(obs_lats)])
  }

  exp_interpolated <- Interpolation(exp = exp, lats = exp_lats, lons = exp_lons, target_grid = target_grid,
                                    method_remap = int_method, points = points, source_file = source_file,
                                    lat_dim = lat_dim, lon_dim = lon_dim, method_point_interp = method_point_interp,
                                    region = region)

  # compute ensemble mean anomalies
  exp_anom <- get_ens_mean_anom(obj_ens = exp_interpolated$data, member_dim = member_dim, sdate_dim = sdate_dim)

  # If after interpolating 'exp' data the coordinates do not match, the obs data is interpolated to
  # the same grid to force the matching
  if ((!.check_coords(lat1 = exp_interpolated$lat, lat2 = obs_lats,
                      lon1 = exp_interpolated$lon, lon2 = obs_lons)) | !is.null(points)) {
    obs_interpolated <- Interpolation(exp = obs, lats = obs_lats, lons = obs_lons, target_grid = target_grid,
                                      method_remap = int_method, points = points, source_file = source_file,
                                      lat_dim = lat_dim, lon_dim = lon_dim,
                                      method_point_interp = method_point_interp, region = region)
    obs_ref <- obs_interpolated$data
  } else {
    obs_ref <- obs
  }

  # convert observations to categorical predictands
  obs_cat <- Apply(obs_ref, target_dims = sdate_dim, function(x) {
                   terc <- convert2prob(as.vector(x), prob = c(1:2/3)) 
                   apply(terc, 1, function(r) which (r == 1))},          
               output_dims = sdate_dim)$output1

  res <- Apply(list(exp_anom, obs_cat), target_dims = sdate_dim, fun = function(x, y)
               .log_reg(x = x, y = y, loocv = loocv), output_dims = sdate_dim)$output1

  res <- list(data = res, lon = exp_interpolated$lon, lat = exp_interpolated$lat)
  
  return(res)
}


get_ens_mean_anom <- function(obj_ens, member_dim, sdate_dim) {

  # compute climatology
  clim <- Apply(obj_ens, target_dims = c(member_dim, sdate_dim), mean)$output1

  # compute ensemble mean
  ens_mean <- Apply(obj_ens, target_dims = member_dim, mean)$output1

  # compute ensemble mean anomalies
  anom <- Apply(list(ens_mean, clim), margins = list(sdate_dim, NULL), fun = function(x,y) x - y)$output1

  return(anom)
}

# atomic functions for logistic regressions
.log_reg <- function(x, y, loocv) {
  
  tmp_df <- data.frame(x = x, y = y)
  
  # if the data is all NA, force return return NA
  if (all(is.na(tmp_df)) | (sum(apply(tmp_df, 2, function(x) !all(is.na(x)))) == 1)) {

    n <- nrow(tmp_df)
    res <- rep(NA, n)

  } else {
    # training
    lm1 <- train_lr(df = tmp_df, loocv = loocv)

    # prediction
    res <- pred_lr(lm1 = lm1, df = tmp_df, loocv = loocv)
  }

  return(res)
}

#-----------------------------------
# Function to train the logistic regressions.
#-----------------------------------
train_lr <- function(df, loocv) {

  require(nnet)
  
  # Remove columns containing only NA's
  df <- df[ , apply(df, 2, function(x) !all(is.na(x)))]

  if (loocv) {

    #lm1 <- lapply(1:nrow(df), function(j) glm(df[-j,], formula = y ~ ., family = "binomial"))
    lm1 <- lapply(1:nrow(df), function(j) multinom(y ~ ., data = df[-j,]))
  } else {

    #lm1 <- list(glm(data = df, formula = y ~ ., family = "binomial"))
    lm1 <- list(multinom(y ~ ., data = df))
  }

  return(lm1)
}

#-----------------------------------
# Function to apply the logistic regressions.
#-----------------------------------
pred_lr <- function(df, lm1, loocv) {

  if (loocv) {

    pred_vals <- sapply(1:nrow(df), function(j) predict(lm1[[j]], df[j,], type = "class"))

  } else {

    pred_vals_ls <- lapply(lm1, predict, data = df, type = "class")
    pred_vals <- unlist(pred_vals_ls)
  }

  return(pred_vals)
}


